from core.models import ObjectType
from dcim.models import Device, Interface
from extras.models import CustomField
from virtualization.models import VirtualMachine, VMInterface


def ensure_custom_fields_exist():
    base_defaults = {
        "required": False,
        "is_cloneable": False,
        "default": None,
        "weight": 100,
        "search_weight": 100,
        "filter_logic": "exact",
        "ui_visible": "always",
        "ui_editable": "yes",
        "group_name": "SmartOS",
    }

    smartos_device_uuid, _ = CustomField.objects.get_or_create(
        name="smartos_device_uuid",
        defaults={
            **base_defaults,
            "label": "UUID",
            "type": "text",
        },
    )
    smartos_device_uuid.object_types.add(ObjectType.objects.get_for_model(Device))

    smartos_version, _ = CustomField.objects.get_or_create(
        name="smartos_version",
        defaults={
            **base_defaults,
            "label": "Version",
            "type": "text",
        },
    )
    smartos_version.object_types.add(ObjectType.objects.get_for_model(Device))

    smartos_nictag, _ = CustomField.objects.get_or_create(
        name="smartos_nictag",
        defaults={
            **base_defaults,
            "label": "Nic Tag",
            "type": "text",
        },
    )
    smartos_nictag.object_types.add(ObjectType.objects.get_for_model(Interface))
    smartos_nictag.object_types.add(ObjectType.objects.get_for_model(VMInterface))

    smartos_brand, _ = CustomField.objects.get_or_create(
        name="smartos_brand",
        defaults={
            **base_defaults,
            "label": "Brand",
            "type": "text",
            "weight": 100,
        },
    )
    smartos_brand.object_types.add(ObjectType.objects.get_for_model(VirtualMachine))

    smartos_image_uuid, _ = CustomField.objects.get_or_create(
        name="smartos_image_uuid",
        defaults={
            **base_defaults,
            "label": "Image UUID",
            "type": "text",
            "weight": 101,
        },
    )
    smartos_image_uuid.object_types.add(
        ObjectType.objects.get_for_model(VirtualMachine)
    )

    smartos_image_name, _ = CustomField.objects.get_or_create(
        name="smartos_image_name",
        defaults={
            **base_defaults,
            "label": "Image Name",
            "type": "text",
            "weight": 102,
        },
    )
    smartos_image_name.object_types.add(
        ObjectType.objects.get_for_model(VirtualMachine)
    )

    smartos_image_version, _ = CustomField.objects.get_or_create(
        name="smartos_image_version",
        defaults={
            **base_defaults,
            "label": "Image Version",
            "type": "text",
            "weight": 103,
        },
    )
    smartos_image_version.object_types.add(
        ObjectType.objects.get_for_model(VirtualMachine)
    )

    smartos_owner_uuid, _ = CustomField.objects.get_or_create(
        name="smartos_owner_uuid",
        defaults={
            **base_defaults,
            "label": "Owner UUID",
            "type": "text",
            "weight": 105,
        },
    )
    smartos_owner_uuid.object_types.add(
        ObjectType.objects.get_for_model(VirtualMachine)
    )

    smartos_billing_uuid, _ = CustomField.objects.get_or_create(
        name="smartos_billing_uuid",
        defaults={
            **base_defaults,
            "label": "Billing UUID",
            "type": "text",
            "weight": 104,
        },
    )
    smartos_billing_uuid.object_types.add(
        ObjectType.objects.get_for_model(VirtualMachine)
    )



import uuid
from contextlib import ExitStack

from dcim.choices import InterfaceTypeChoices
from dcim.fields import mac_unix_expanded_uppercase
from dcim.models import MACAddress
from django.db import transaction
from django.utils import timezone
from ipam.models import IPAddress
from netaddr import EUI, IPNetwork
from netbox.jobs import JobRunner
from netbox.registry import registry
from users.models import User
from utilities.request import NetBoxFakeRequest
from virtualization.choices import VirtualMachineStatusChoices
from virtualization.models import VirtualMachine

from netbox_smartos.custom_fields import ensure_custom_fields_exist

VM_STATE_MAP = {
    "configured": VirtualMachineStatusChoices.STATUS_PLANNED,
    "down": VirtualMachineStatusChoices.STATUS_OFFLINE,
    "failed": VirtualMachineStatusChoices.STATUS_FAILED,
    "incomplete": VirtualMachineStatusChoices.STATUS_PLANNED,
    "installed": VirtualMachineStatusChoices.STATUS_PLANNED,
    "provisioning": VirtualMachineStatusChoices.STATUS_PLANNED,
    "ready": VirtualMachineStatusChoices.STATUS_PLANNED,
    "receiving": VirtualMachineStatusChoices.STATUS_PLANNED,
    "running": VirtualMachineStatusChoices.STATUS_ACTIVE,
    "shutting_down": VirtualMachineStatusChoices.STATUS_DECOMMISSIONING,
    "stopped": VirtualMachineStatusChoices.STATUS_OFFLINE,
    "stopping": VirtualMachineStatusChoices.STATUS_DECOMMISSIONING,
}


def update_attr(obj, attr, value):
    if not value:
        return

    if getattr(obj, attr) == value:
        return

    setattr(obj, attr, value)
    obj.has_changed = True


def update_cf_attr(obj, attr, value):
    if not value:
        return

    if obj.custom_field_data.get(attr) == value:
        return

    obj.custom_field_data[attr] = value
    obj.has_changed = True


def discover_mac(interface, mac_address):
    if not mac_address:
        return

    mac_changed = False
    if not interface.primary_mac_address:
        if getattr(interface, "has_changed", False) or interface.pk is None:
            interface.save()
        interface.primary_mac_address = MACAddress(assigned_object=interface)
        interface.has_changed = True
        mac_changed = True

    formatted_mac = EUI(mac_address, version=48, dialect=mac_unix_expanded_uppercase)
    if interface.primary_mac_address.mac_address != formatted_mac:
        interface.primary_mac_address.mac_address = formatted_mac
        mac_changed = True
    if mac_changed:
        interface.primary_mac_address.save()


class ProcessSmartOSReportJob(JobRunner):
    class Meta:
        name = "Process SmartOS Report"

    def run(self, device, *args, **kwargs):
        self.logger.debug(f"Processing {device}")

        user, _ = User.objects.get_or_create(username="smartos-bot")
        request = NetBoxFakeRequest(
            {
                "META": {},
                "POST": {},
                "GET": {},
                "FILES": {},
                "user": user,
                "path": "",
                "id": uuid.uuid4(),
            }
        )

        with ExitStack() as stack:
            for request_processor in registry["request_processors"]:
                stack.enter_context(request_processor(request))

            with transaction.atomic():
                self._process(device)

    def _process(self, device):
        ensure_custom_fields_exist()
        sysinfo = device.smartos_report.report.get("sysinfo", {})
        vms = device.smartos_report.report.get("vm", {})
        imgs = device.smartos_report.report.get("img", {})
        img_by_uuid = {
            img["manifest"]["uuid"]: img
            for img in imgs
            if "manifest" in img and "uuid" in img["manifest"]
        }

        device.has_changed = False

        update_attr(device, "serial", sysinfo.get("Serial Number"))

        update_cf_attr(device, "smartos_device_uuid", sysinfo.get("UUID"))
        update_cf_attr(device, "smartos_version", sysinfo.get("Live Image"))

        if device.has_changed:
            device.save()

        # Interfaces
        nics = sysinfo.get("Network Interfaces", {})
        for name, nic_info in nics.items():
            interface, _ = device.interfaces.get_or_create(name=name)

            interface.has_changed = False
            mac_address = nic_info.get("MAC Address")
            discover_mac(interface, mac_address)

            nic_tags = ",".join(sorted(nic_info.get("NIC Names", [])))
            update_cf_attr(interface, "smartos_nictag", nic_tags)

            if interface.has_changed:
                interface.save()

            # IPv4 can't be done bacause we don't have the prefix len

        # Nictags in "Virtual Network Interfaces"
        vnics = sysinfo.get("Virtual Network Interfaces", {})
        for name, nic_info in vnics.items():
            interface, _ = device.interfaces.get_or_create(name=name)
            update_attr(interface, "type", InterfaceTypeChoices.TYPE_VIRTUAL)

            interface.has_changed = False
            mac_address = nic_info.get("MAC Address")
            discover_mac(interface, mac_address)

            if parent_name := nic_info.get("Host Interface"):
                parent = device.interfaces.filter(name=parent_name).first()
                update_attr(interface, "parent", parent)

            if interface.has_changed:
                interface.save()

        # VMs
        vm_ids = set([vm["uuid"] for vm in vms])
        device.virtual_machines.exclude(name__in=vm_ids).delete()
        for vm_info in vms:
            if (
                VirtualMachine.objects.exclude(device=device)
                .filter(name=vm_info["uuid"])
                .exists()
            ):
                self.logger.warning(f"Duplicate VM with UUID {vm_info['uuid']}")
                continue

            vm, _ = VirtualMachine.objects.get_or_create(name=vm_info["uuid"])
            vm.has_changed = False

            update_attr(vm, "description", vm_info.get("alias"))
            update_attr(vm, "device", device)
            update_attr(vm, "cluster", device.cluster)
            update_attr(vm, "site", device.site)
            update_attr(vm, "memory", vm_info.get("ram"))
            update_attr(vm, "disk", vm_info.get("quota", 0) * 1024)
            update_attr(vm, "vcpus", vm_info.get("cpu_cap", 0) / 100)

            state = vm_info.get("state")
            if state in VM_STATE_MAP:
                update_attr(vm, "status", VM_STATE_MAP[state])

            update_cf_attr(vm, "smartos_brand", vm_info.get("brand"))
            update_cf_attr(vm, "smartos_owner_uuid", vm_info.get("owner_uuid"))
            update_cf_attr(vm, "smartos_billing_uuid", vm_info.get("billing_id"))

            img_uuid = vm_info.get("image_uuid")
            img_manifest = img_by_uuid.get(img_uuid, {}).get("manifest", {})
            update_cf_attr(vm, "smartos_image_uuid", img_uuid)
            update_cf_attr(vm, "smartos_image_name", img_manifest.get("name"))
            update_cf_attr(vm, "smartos_image_version", img_manifest.get("version"))

            if vm.has_changed:
                vm.save()

            nic_by_name = {nic["interface"]: nic for nic in vm_info.get("nics", [])}
            vm.interfaces.exclude(name__in=nic_by_name.keys()).delete()
            for name, nic_info in nic_by_name.items():
                nic, _ = vm.interfaces.get_or_create(name=name)

                nic.has_changed = False
                discover_mac(nic, nic_info.get("mac"))
                update_cf_attr(nic, "smartos_nictag", nic_info.get("nic_tag"))

                if nic.has_changed:
                    nic.save()

                addrs = []
                for addr in nic_info.get("ips", []):
                    if addr in ["dhcp", "addrconf"]:
                        continue
                    ip_addr = IPNetwork(addr)

                    addr = (
                        IPAddress.objects.filter(
                            address__net_host_contained=ip_addr.ip
                        ).first()
                        or IPAddress()
                    )
                    if addr.pk:
                        addr.has_changed = False
                    else:
                        addr.has_changed = True

                    update_attr(addr, "address", ip_addr)

                    # TODO: if primary nic we could also set this as primary so on vm

                    if addr.has_changed:
                        addr.save()

                    addrs += [addr]

                nic.ip_addresses.set(addrs)

        device.smartos_report.last_processed = timezone.now()
        device.smartos_report.save()



from dcim.api.serializers import DeviceSerializer
from rest_framework import serializers

from netbox_smartos.models import SmartOSReport


class SmartOSReportSerializer(serializers.Serializer):
    device = DeviceSerializer(nested=True, read_only=True)
    token = serializers.CharField(write_only=True)
    report = serializers.JSONField()

    class Meta:
        model = SmartOSReport
        fields = ("id", "device", "token", "report", "last_heartbeat", "last_processed")
        brief_fields = ("id", "device")



import logging

from dcim.models import Device
from django.utils import timezone
from netbox.plugins import get_plugin_config
from rest_framework.exceptions import AuthenticationFailed, ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView

from netbox_smartos.api.serializers import SmartOSReportSerializer
from netbox_smartos.jobs import ProcessSmartOSReportJob

logger = logging.getLogger("netbox_smartos.api.views")


class DeviceReportView(APIView):
    permission_classes = []
    serializer_class = SmartOSReportSerializer
    queryset = Device.objects.all()

    def post(self, request):
        serializer = SmartOSReportSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        token = serializer.validated_data["token"]
        if not token or (token != get_plugin_config("netbox_smartos", "api_token")):
            raise AuthenticationFailed()

        report = serializer.validated_data["report"]
        device_uuid = report.get("sysinfo", {}).get("UUID", "")
        device_hostname = report.get("sysinfo", {}).get("Hostname", "")

        device = None
        if device_uuid:
            device = Device.objects.filter(
                custom_field_data__smartos_device_uuid=device_uuid
            ).first()
        if not device and device_hostname:
            device = Device.objects.filter(name=device_hostname).first()

        if not device:
            raise ValidationError({"device": "Could not find matching device."})

        device.smartos_report.report = report
        device.smartos_report.last_heartbeat = timezone.now()
        device.smartos_report.save()

        ProcessSmartOSReportJob.enqueue(device=device)

        return Response({"status": "ok"})



from annoying.fields import AutoOneToOneField
from dcim.models import Device
from django.db import models


class SmartOSReport(models.Model):
    device = AutoOneToOneField(
        to=Device, on_delete=models.CASCADE, related_name="smartos_report"
    )
    report = models.JSONField(default=dict)
    last_heartbeat = models.DateTimeField(null=True)
    last_processed = models.DateTimeField(null=True)

    def __str__(self):
        return self.device.name



# Generated by Django 5.2.8 on 2025-11-24 15:29

import annoying.fields
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("dcim", "0215_rackreservation_status"),
    ]

    operations = [
        migrations.CreateModel(
            name="SmartOSReport",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("report", models.JSONField(default=dict)),
                ("last_heartbeat", models.DateTimeField(null=True)),
                ("last_processed", models.DateTimeField(null=True)),
                (
                    "device",
                    annoying.fields.AutoOneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="smartos_report",
                        to="dcim.device",
                    ),
                ),
            ],
        ),
    ]



from importlib.metadata import metadata

from netbox.plugins import PluginConfig

metadata = metadata("netbox_smartos")


class NetboxSmartOSConfig(PluginConfig):
    name = "netbox_smartos"
    verbose_name = "Netbox SmartOS"
    description = "SmartOS integration"
    version = metadata.get("Version")
    base_url = "smartos"
    min_version = "4.2.0"
    django_apps = []
    author = "SkyLime GmbH"
    author_email = "info@skylime.net"


config = NetboxSmartOSConfig



from dcim.models.sites import Site
from django.contrib.auth.hashers import make_password
from django.core.management.base import BaseCommand
from users.models import User

from netbox_smartos.custom_fields import ensure_custom_fields_exist


def create_admin():
    admin, _ = User.objects.get_or_create(
        username="admin",
        defaults={
            "email": "admin@example.com",
            "first_name": "Super",
            "last_name": "User",
            "is_staff": True,
            "is_superuser": True,
            "is_active": True,
        },
    )
    admin.password = make_password("admin")
    admin.save()
    return admin


def create_initial_site():
    Site.objects.get_or_create(
        name="demo",
        defaults={
            "slug": "demo",
        },
    )


class Command(BaseCommand):
    help = "Create initial data"

    def add_arguments(self, parser):
        parser.add_argument(
            "--dev",
            action="store_true",
            dest="dev",
            default=False,
            help="Also create dev seed data",
        )

    def handle(self, *args, **options):
        if options["dev"]:
            create_admin()
            create_initial_site()
        ensure_custom_fields_exist()
